import { EntryIndex } from "./entry-index.js";
import * as Sorting from "./log-sorting.js";
import { EntryFetchAllOptions, EntryFetchOptions } from "./entry-io.js";
import { EncryptionTemplateMaybeEncrypted, Entry, Payload, CanAppend, EntryType } from "./entry.js";
import { HLC, LamportClock, Timestamp } from "./clock.js";
import { PublicKeyEncryptionResolver, SignatureWithKey } from "@dao-xyz/peerbit-crypto";
import { Encoding } from "./encoding.js";
import { Identity } from "./identity.js";
import { HeadsIndex } from "./heads.js";
import { BlockStore } from "@dao-xyz/libp2p-direct-block";
import { Values } from "./values.js";
import { TrimOptions } from "./trim.js";
export type Change<T> = {
    added: Entry<T>[];
    removed: Entry<T>[];
};
export type Change2<T> = {
    added: Entry<T>;
    removed: Entry<T>[];
};
export type LogOptions<T> = {
    encryption?: PublicKeyEncryptionResolver;
    encoding?: Encoding<T>;
    logId?: string;
    clock?: LamportClock;
    sortFn?: Sorting.ISortFunction;
    concurrency?: number;
    trim?: TrimOptions;
};
/**
 * @description
 * Log implements a G-Set CRDT and adds ordering.
 *
 * From:
 * "A comprehensive study of Convergent and Commutative Replicated Data Types"
 * https://hal.inria.fr/inria-00555588
 */
export declare class Log<T> {
    private _sortFn;
    private _storage;
    private _id;
    private _hlc;
    private _identity;
    private _entryIndex;
    private _headsIndex;
    private _values;
    private _nextsIndex;
    private _encryption?;
    private _encoding;
    private _trim;
    private _entryCache;
    constructor(store: BlockStore, identity: Identity, options?: LogOptions<T>);
    get id(): string;
    /**
     * Returns the length of the log.
     */
    get length(): number;
    get values(): Values<T>;
    /**
     * Checks if a entry is part of the log
     * @param {string} hash The hash of the entry
     * @returns {boolean}
     */
    has(cid: string): boolean;
    /**
     * Get all entries sorted. Don't use this method anywhere where performance matters
     */
    toArray(): Promise<Entry<T>[]>;
    /**
     * Returns the head index
     */
    get headsIndex(): HeadsIndex<T>;
    /**
     * Don't use this anywhere performance matters
     */
    getHeads(): Promise<Entry<T>[]>;
    /**
     * Returns an array of Entry objects that reference entries which
     * are not in the log currently.
     * @returns {Array<Entry<T>>}
     */
    getTails(): Promise<Entry<T>[]>;
    /**
     * Returns an array of hashes that are referenced by entries which
     * are not in the log currently.
     * @returns {Array<string>} Array of hashes
     */
    getTailHashes(): Promise<string[]>;
    /**
     * Get local HLC
     */
    get hlc(): HLC;
    get identity(): Identity;
    get storage(): BlockStore;
    get nextsIndex(): Map<string, Set<string>>;
    get entryIndex(): EntryIndex<T>;
    get encryption(): PublicKeyEncryptionResolver | undefined;
    get encoding(): Encoding<T>;
    get sortFn(): Sorting.ISortFunction;
    /**
     * Set the identity for the log
     * @param {Identity} [identity] The identity to be set
     */
    setIdentity(identity: Identity): void;
    /**
     * Find an entry.
     * @param {string} [hash] The hashes of the entry
     */
    get(hash: string): Promise<Entry<T> | undefined>;
    traverse(rootEntries: Entry<T>[], amount?: number, endHash?: string): Promise<{
        [key: string]: Entry<T>;
    }>;
    getReferenceSamples(from: Entry<T>, options?: {
        pointerCount?: number;
        memoryLimit?: number;
    }): Promise<Entry<T>[]>;
    /**
     * Append an entry to the log.
     * @param {Entry} entry Entry to add
     * @return {Log} New Log containing the appended value
     */
    append(data: T, options?: {
        type?: EntryType;
        canAppend?: CanAppend<T>;
        gidSeed?: Uint8Array;
        nexts?: Entry<any>[];
        identity?: Identity;
        signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
        reciever?: EncryptionTemplateMaybeEncrypted;
        onGidsShadowed?: (gids: string[]) => void;
        trim?: TrimOptions;
        timestamp?: Timestamp;
    }): Promise<{
        entry: Entry<T>;
        removed: Entry<T>[];
    }>;
    reset(entries: Entry<T>[], heads?: Entry<T>[]): Promise<void>;
    iterator(options?: {
        from?: "tail" | "head";
        amount?: number;
    }): IterableIterator<string>;
    trim(option?: TrimOptions | undefined): Promise<Entry<T>[]>;
    _joining: Map<string, Promise<any>>;
    join(entriesOrLog: (string | Entry<T>)[] | Log<T>, options?: {
        canAppend?: (entry: Entry<T>) => Promise<boolean> | boolean;
        onChange?: (change: Change2<T>) => void | Promise<void>;
        verifySignatures?: boolean;
        trim?: TrimOptions;
    }): Promise<void>;
    private joinEntry;
    private processEntry;
    deleteRecursively(from: Entry<any> | Entry<any>[], skipFirst?: boolean): Promise<Entry<T>[]>;
    delete(entry: Entry<any>): Promise<void>;
    /**
     * Get the log in JSON format.
     * @returns {Object} An object with the id and heads properties
     */
    toJSON(): Promise<{
        id: string;
        heads: string[];
    }>;
    /**
     * Get the log in JSON format as a snapshot.
     * @returns {Object} An object with the id, heads and value properties
     */
    toSnapshot(): {
        id: string;
        heads: string[];
        values: Promise<Entry<T>[]>;
    };
    /**
     * Returns the log entries as a formatted string.
     * @returns {string}
     * @example
     * two
     * └─one
     *   └─three
     */
    toString(payloadMapper?: (payload: Payload<T>) => string): Promise<string>;
    /**
     * Get the log's multihash.
     * @returns {Promise<string>} Multihash of the Log as Base58 encoded string.
     */
    toMultihash(options?: {
        format?: string;
    }): Promise<string>;
    static fromMultihash<T>(store: BlockStore, identity: Identity, hash: string, options: {
        sortFn?: Sorting.ISortFunction;
    } & EntryFetchAllOptions<T>): Promise<Log<T>>;
    static fromEntryHash<T>(store: BlockStore, identity: Identity, hash: string | string[], options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        logId?: string;
        length?: number;
        exclude?: any[];
        shouldFetch?: (hash: string) => boolean;
        timeout?: number;
        concurrency?: number;
        sortFn?: any;
        onFetched?: any;
        replicate?: boolean;
    }): Promise<Log<T>>;
    /**
     * Create a log from a Log Snapshot JSON.
     * @param {IPFS} ipfs An IPFS instance
     * @param {Identity} identity The identity instance
     * @param {Object} json Log snapshot as JSON object
     * @param {Object} options
     * @param {AccessController} options.access The access controller instance
     * @param {number} options.length How many entries to include in the log
     * @param {function(hash, entry,  parent, depth)} [options.onFetched]
     * @param {Function} options.sortFn The sort function - by default LastWriteWins
     * @return {Promise<Log>} New Log
     */
    static fromJSON<T>(store: BlockStore, identity: Identity, json: {
        id: string;
        heads: string[];
    }, options?: {
        encoding?: Encoding<T>;
        encryption?: PublicKeyEncryptionResolver;
        length?: number;
        timeout?: number;
        sortFn?: Sorting.ISortFunction;
        onFetched?: (entry: Entry<T>) => void;
        replicate?: boolean;
    }): Promise<Log<T>>;
    static fromEntry<T>(store: BlockStore, identity: Identity, sourceEntries: Entry<T>[] | Entry<T>, options: EntryFetchOptions<T> & {
        shouldFetch?: (hash: string) => boolean;
        encryption?: PublicKeyEncryptionResolver;
        sortFn?: Sorting.ISortFunction;
    }): Promise<Log<T>>;
    /**
     * Find heads from a collection of entries.
     *
     * Finds entries that are the heads of this collection,
     * ie. entries that are not referenced by other entries.
     *
     * @param {Array<Entry<T>>} entries Entries to search heads from
     * @returns {Array<Entry<T>>}
     */
    static findHeads<T>(entries: Entry<T>[]): Entry<T>[];
    static findTails<T>(entries: Entry<T>[]): Entry<T>[];
    static findTailHashes(entries: Entry<any>[]): string[];
}
